import sys

import cloup

from .. import answers, jinja, layout, questions, template
from .cli import cli


def validate_template_file(_ctx, _param, value):
    if value.startswith("/"):
        sys.exit(f'error: template filename "{value}" must be relative path to template directory')
    return value


#
# mk-scaffold clone [OPTIONS] TEMPLATE
#
@cli.command(help="Clone and templatize a repository.")
@cloup.option(
    "-i",
    "--input-file",
    "answer_file",
    metavar="FILE",
    type=cloup.Path(exists=True, dir_okay=False, readable=True),
    help="Location of a yaml input file, usually named '.answers.yml', with answers to the questions.",
)
@cloup.option(
    "-o",
    "--output-dir",
    metavar="DIR",
    default=".",
    type=cloup.Path(file_okay=False, dir_okay=True),
    show_default=False,
    help="Where to output the generated files. [default: current directory]",
)
@cloup.option(
    "--filename",
    "template_filename",
    metavar="FILENAME",
    default="scaffold.yml",
    show_default=True,
    callback=validate_template_file,
    help="Filename of the scaffold file to use.",
)
@cloup.option(
    "-b",
    "--branch",
    metavar="BRANCH",
    default=None,
    show_default=False,
    help="Checkout git BRANCH of git repository",
)
@cloup.argument(
    "template_path",
    nargs=1,
    metavar="TEMPLATE",
    type=cloup.Path(),
    help="Directory or git repository that contains 'scaffold.yml' template file",
)
def clone(**kwargs):
    # Load the answers with a default jinja
    env, ctx = jinja.create()

    answer = answers.load(kwargs.get("answer_file"))
    answer = answers.validate(answer)
    ctx = answers.process(answer, env, ctx)

    tpl = template.Template(**kwargs)
    tpl = template.find(tpl)
    tpl = template.load(tpl)

    tpl, env, ctx = questions.prompt(tpl, ctx)

    tpl = layout.clone(tpl, env, ctx, **kwargs)
    tpl = layout.process(tpl, env, ctx, **kwargs)



from prompt_toolkit.validation import ValidationError

from ... import utils


def validate(value, schema):
    """
    Return { value to use, stop (True/False) }

    If stop is True, the rest of the validators are skipped.
    """
    # ctrl-d gives us a None
    if value is None:
        return None, False

    vartype = schema.get("type", "string")
    if vartype == "string":
        value = str(value)

    if vartype == "integer":
        try:
            value = int(value)
        except ValueError:
            raise ValidationError(message="Invalid type. Expected an integer") from None

    if vartype == "boolean":
        try:
            value = utils.string_as_bool(value)
            return value, True
        except ValueError:
            raise ValidationError(message="Invalid type. Expected a boolean") from None

    return value, False



import os
import sys
import urllib


def accept(tpl, dirpath):
    """
    Uses dirpath and not tpl.path
    """
    # Has to be a directory
    if not os.path.isdir(dirpath):
        return False

    # filepath is the path of the scaffold.yml file, and
    # must exist
    filepath = os.path.join(dirpath, tpl.filename)
    if not os.path.isfile(filepath):
        return False

    # Ensure that filepath is relative to the tpl.path
    real_dirpath = os.path.realpath(dirpath)
    real_filepath = os.path.realpath(filepath)
    if not real_filepath.startswith(real_dirpath):
        sys.exit(f'error: File "{tpl.filename}" is not relative to the template path "{dirpath}"')

    return real_dirpath


def find(tpl):
    # A local directory can also be a git repository
    # So if a branch is specified, don't look for a local
    # directory, but return directly to force the git handler
    if tpl.branch:
        return False

    # If a scheme is detected, it's also not a local directory
    urlpath = urllib.parse.urlparse(tpl.path)
    if urlpath.scheme in ["https", "http"]:
        return False

    if os.path.isfile(tpl.path):
        sys.exit(
            "error: positional argument TEMPLATE should point to a directory, or remote git repository, and not a file",
        )

    return accept(tpl, tpl.path)



from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    """
    Return { value to use, stop (True/False) }

    If stop is True, the rest of the validators are skipped.
    """
    vartype = schema.get("type", "string")
    if vartype != "string":
        return value, False

    min_length = schema.get("min_length", None)
    if min_length and len(value) < min_length:
        raise ValidationError(message=f"Answer cannot be shorter than {min_length} characters") from None
    return value, False



from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    """
    Return { value to use, stop (True/False) }

    If stop is True, the rest of the validators are skipped.
    """
    vartype = schema.get("type", "integer")
    if vartype != "integer":
        return value, False

    max_value = schema.get("max_value", None)
    if max_value and int(value) > max_value:
        raise ValidationError(message=f"Answer cannot be more than {max_value}") from None
    return value, False



"""
Ask questions and answer them by prompting the user

Special cases to be handled:
- ctrl-d: Set value to None

"""

import sys
from contextlib import nullcontext

import prompt_toolkit
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.validation import ValidationError, Validator

from .. import jinja, utils
from .validators import allowed, max_length, max_value, min_length, min_value, nullable, vartype

bindings = KeyBindings()


# pylint: disable=redefined-outer-name
def _stdin_input(prompt):
    """
    Convenience function in order to mock the builtin
    """
    return input(prompt)


def _validate(value, schema):
    """
    Validate the answer and return modified answer if needed.

    Return value + stop (True/False)
    """
    # Order matters. By checking nullable and default first
    # we can make assumptions on values (null? not null? etc.)
    validators = [
        vartype,
        nullable,
        allowed,
        max_length,
        min_length,
        min_value,
        max_value,
    ]

    # Returns (value, True) if we are to stop iterating,
    # we are to replace value by a new value
    for validator in validators:
        value, stop = validator.validate(value, schema)
        if stop is True:
            return value
    return value


def _prepare(question, env, ctx):
    """
    Determine if the question is to be asked, and
    if so, build a prompt

    Return true if question is to be asked.
    """

    # Prepare all values in order to have an json snapshot of
    # every question
    name = question["name"]

    # Schema might not be present by default
    schema = question.get("schema", {})
    question["schema"] = schema

    prompt = name
    prompt = str(utils.eval_string(prompt, env, ctx))
    prompt += ": "
    question["prompt"] = prompt

    description = question.get("description")
    question["description"] = str(utils.eval_string(description, env, ctx))

    hidden = question.get("hidden", False)
    question["hidden"] = utils.eval_string(hidden, env, ctx)

    # Will we prompt this question?
    condition = question.get("if")
    condition = utils.eval_string(condition, env, ctx)
    if condition is not None:
        question["if"] = condition
        if not condition:
            return False

    # Set the answer as the "default", override if necessary
    # any previous "default" value
    # It will be eval'd during the "default" code block
    answer = ctx["scaffold"].get(question["name"])
    if answer is not None:
        schema["default"] = answer

    # Prepare the default value
    default = schema.get("default")
    default = utils.eval_string(default, env, ctx)
    if default is None:
        schema.pop("default", None)
    else:
        schema["default"] = default
    return True


def prompt(tpl, ctx):
    """
    For every question in the input file, ask the question
    and record the answer in the context
    """
    env, _ = jinja.create(tpl.data["jinja2"])

    questions = tpl.data["questions"]
    for question in questions:
        if not _prepare(question, env, ctx):
            continue

        name, value = _prompt(question)

        question["value"] = value
        ctx = jinja.ctx_add(ctx, name, value)

    return tpl, env, ctx


def _prompt(question):
    """
    Ask the question until it is answered or canceled

    Returns key, value
    """
    name = question.get("name")
    prompt = question.get("prompt")
    hidden = question.get("hidden")
    description = question.get("description")
    schema = question.get("schema")
    default = schema.get("default")

    def prevalidate(x, schema):
        # Exceptions are raised, always return True
        _validate(x, schema)
        return True

    validator = Validator.from_callable(lambda x: prevalidate(x, schema))

    while True:
        try:
            # if hidden, then return default, which was previously set to
            # the value in answer if present, and otherwise
            # to the default from scaffold.yml
            if hidden:
                answer = default
            elif sys.stdin.isatty():
                kwargs = {}
                if default is not None:
                    kwargs["default"] = str(default)

                answer = prompt_toolkit.prompt(
                    prompt,
                    validator=validator,
                    bottom_toolbar=description,
                    key_bindings=bindings,
                    validate_while_typing=False,
                    **kwargs,
                )
            else:
                answer = _stdin_input(prompt)
            answer = _validate(answer, schema)
            return name, answer

        except EOFError:
            # ctrl-d was used
            try:
                answer = _validate(None, schema)
                return name, answer
            except ValidationError:
                continue



# vim: foldmarker=[[,]] foldmethod=marker
import cerberus

# https://docs.python-cerberus.org/en/stable/validation-rules.html
SCHEMA = r"""
---
answers: # [[
  type: dict
  default: {}
  nullable: true
  keysrules:
    type: string
    regex: '[a-zA-Z0-9_]+'
# ]]
"""


class LocalValidator(cerberus.Validator):
    pass



from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    """
    Return { value to use, stop (True/False) }

    If stop is True, the rest of the validators are skipped.
    """
    allowed = schema.get("allowed")
    if allowed is None:
        return value, False

    if value not in allowed:
        allowed = ", ".join(allowed)
        raise ValidationError(message=f"Answer must be within [{allowed}]")

    return value, False



import atexit
import os
import shutil
import subprocess
import sys
import tempfile

from . import directory


def _get_git():
    retval = shutil.which("git")
    if retval is None:
        print('warning: "git" executable was not found', file=sys.stderr)
    return retval


def _get_tmpdir():
    """
    Create a temporary folder to be deleted at exit
    """
    tmpdir = tempfile.mkdtemp(prefix="scaffold-git-")

    def remove_tmpdir():
        shutil.rmtree(tmpdir)

    atexit.register(remove_tmpdir)
    return tmpdir


def _clone(tpl):
    git = _get_git()
    tmpdir = _get_tmpdir()

    path = tpl.path
    cmdline = [git, "clone", "--single-branch"]
    if tpl.branch:
        cmdline += ["--branch", tpl.branch]
    if os.path.exists(tpl.path) and os.path.isdir(tpl.path):
        # This is a local repository and git prefers with the
        # "file://"
        path = f"file://{tpl.path}"
    else:
        cmdline += ["--depth", "1"]
    cmdline += [path, "repository"]

    try:
        subprocess.run(cmdline, cwd=tmpdir, check=True)
    except subprocess.CalledProcessError as err:
        sys.exit(f'error: failed to clone git repository "{tpl.path}": {err}')

    dirpath = os.path.join(tmpdir, "repository")
    return directory.accept(tpl, dirpath)


def find(tpl):
    """
    path can be a remote git+ssh, or git+http, or even a local path
    if branch is None, we checkout the default branch, usually main or master
    """
    git = _get_git()
    if not git:
        return False
    return _clone(tpl)



import copy
import json
import os
import sys

import yaml

from .. import constants as c
from . import schema
from .methods import directory, git


class Template:
    def __init__(self, **kwargs):
        self.path = kwargs.get("template_path")
        self.filename = kwargs.get("template_filename", c.TEMPLATE_FILENAME)
        self.branch = kwargs.get("branch")
        self.workpath = kwargs.get("workpath")
        self.includes = None
        self.data = None

    def __repr__(self):
        return json.dumps(
            {
                "path": self.path,
                "filename": self.filename,
                "branch": self.branch,
                "workpath": self.workpath,
            },
        )

    @property
    def actions(self):
        if self.data is None:
            return []
        # pylint: disable=unsubscriptable-object
        return self.data["actions"]

    @property
    def children(self):
        """
        Return only the includes for this template
        """
        if not self.data:
            return

        # By reversing the list, we'll be building the list
        # of questions in the right order
        items = reversed(self.data.get("inherit", []))
        for item in items:
            path = item["include"]
            filename = item.get("filename")
            branch = item.get("branch")
            workpath = item.get("workpath")

            yield Template(template_path=path, template_filename=filename, branch=branch, workpath=workpath)

    @property
    def dependencies(self):
        # By reversing the list, we'll be building the list
        # of questions in the right order
        items = reversed(self._dependencies)
        yield from items


def find(tpl):
    for func in [directory.find, git.find]:
        tpl.workpath = func(tpl)
        if tpl.workpath:
            return tpl
    sys.exit(f'error: no "{tpl.filename}" file found in "{tpl.path}"')


def _load(tpl):
    path = os.path.join(tpl.workpath, tpl.filename)
    try:
        with open(path, encoding="UTF-8") as fd:
            data = yaml.safe_load(fd) or {}
    except Exception as err:
        sys.exit(f"error: failed to open '{tpl.workpath}': {err}")
    return data


def _reduce(data):
    # Reduce to relevant sections.
    # This allows a user to add whatever he wants to a scaffold.yml file
    retval = {
        "actions": data.get("actions", []),
        "answers": data.get("answers", {}),
        "inherit": data.get("inherit", []),
        "jinja2": data.get("jinja2", {}),
        "questions": data.get("questions", []),
    }
    return retval


def _validate(data):
    # TODO: merge with answers into utils
    yaml_schema = yaml.safe_load(schema.SCHEMA)
    validator = schema.LocalValidator(yaml_schema)

    if not validator.validate(data):
        locations = str(json.dumps(validator.errors, indent=2))
        raise SystemExit(f"error: YAML schema validation error. Location:\n{locations}") from None

    return validator.normalized(data)


def load(tpl, root=None):
    if root is None:
        root = tpl

    data = _load(tpl)
    data = _reduce(data)
    data = _validate(data)
    tpl.data = data

    # Keep a list of all encountered children
    if root.includes is None:
        root.includes = []
    root.includes += [tpl]

    # In order to load the questions, we need to recurse into
    # all files and build the questions list by priority, the
    # inclusion order
    return _recurse(tpl, root)


def _recurse(tpl, root):
    for other in tpl.children:
        other = find(other)
        other = load(other, root)
        tpl = _merge(tpl, other)
    return tpl


def _merge(tpl, other):
    # Merge questions.
    # Since the "name" must be unique, we use it as a key
    # other has priority on self
    result = {i["name"]: i for i in other.data["questions"]}
    for question in tpl.data["questions"]:
        result[question["name"]] = question
    tpl.data["questions"] = [v for k, v in result.items()]

    # Sort by order if present
    tpl.data["questions"] = sorted(tpl.data["questions"], key=lambda item: item.get("order", 0))

    return tpl



import copy
import json
import sys

import yaml

from .. import constants as c
from .. import jinja, utils
from . import schema


def load(filepath):
    if not filepath:
        return None

    try:
        with open(filepath, encoding="UTF-8") as fd:
            data = yaml.safe_load(fd)
    except Exception as err:
        sys.exit(f"error: failed to open '{filepath}': {err}")

    if data is None:
        data = {}
    data = {"answers": data.get("answers")}

    return data


def validate(data):
    yaml_schema = yaml.safe_load(schema.SCHEMA)
    validator = schema.LocalValidator(yaml_schema)

    if data is None:
        data = {}
    else:

        if not validator.validate(data):
            locations = str(json.dumps(validator.errors, indent=2))
            raise SystemExit(f"error: YAML answer file schema validation error. Location:\n{locations}") from None

    return validator.normalized(data)


def process(data, env, ctx):
    if data is None:
        return ctx

    answers = data.get("answers", {})
    if answers is None:
        return ctx

    for k, v in answers.items():
        v = utils.eval_string(v, env, ctx)
        ctx = jinja.ctx_add(ctx, k, v)
    return ctx



from datetime import datetime
from pathlib import Path

from .environment import StrictNativeEnvironment


def create(opts=None):
    opts = opts or {}
    ctx = {
        "scaffold": {},
        "year": datetime.now().year,
        "curdir": str(Path.cwd()),
    }
    env = StrictNativeEnvironment(**opts)
    return env, ctx


def ctx_add(ctx, key, value):
    ctx["scaffold"][key] = value
    return ctx



import os
import shutil


def _if(item, env, ctx):
    retval = item["if"]
    if retval not in [True, False]:
        retval = env.from_string(retval).render(**ctx)
    return retval


def _action_move(item, env, ctx, output_dir):

    src = str(env.from_string(item["src"]).render(**ctx))
    src = os.path.join(output_dir, src)

    dst = str(env.from_string(item["dst"]).render(**ctx))
    dst = os.path.join(output_dir, dst)
    # Ensure path is subpath
    # TODO   if not Path(dst).is_relative_to(output_dir):
    #        print(f'warning: skipping path "{dst_orig}" as it is not relative to output path.', file=sys.stderr)
    #        return

    if _if(item, env, ctx):
        if os.path.exists(src):
            shutil.move(src, dst)
    elif item.get("else") == "remove":
        # TODO if not Path(src).is_relative_to(output_dir):
        # TODO     print(f'warning: skipping path "{src}" as it is not relative to output path.', file=sys.stderr)
        # TODO     return
        if os.path.exists(src):
            os.remove(src)


def _action_remove(item, env, ctx, output_dir):
    if not _if(item, env, ctx):
        return

    for path in item["path"]:
        path = str(env.from_string(path).render(**ctx))
        path = os.path.join(output_dir, path)

        # Ensure path is subpath
        # TODOif not Path(path).is_relative_to(output_dir):
        # TODO    print(f'warning: skipping path "{path_orig}" as it is not relative to output path.', file=sys.stderr)
        # TODO    continue

        if os.path.islink(path) or os.path.isfile(path):
            os.remove(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)


def _action_remove_nl(item, env, ctx, output_dir):
    if not _if(item, env, ctx):
        return

    for path in item["path"]:
        path = str(env.from_string(path).render(**ctx))
        path = os.path.join(output_dir, path)

        with open(path, encoding="UTF-8") as fd:
            data = fd.read()
        data = data.rstrip()
        with open(path, mode="w", encoding="UTF-8") as fd:
            fd.write(data)


def _process(tpl, env, ctx, output_dir):
    for item in tpl.actions:
        action = item["action"]
        if action == "move":
            _action_move(item, env, ctx, output_dir)
        elif action == "remove":
            _action_remove(item, env, ctx, output_dir)
        elif action == "remove-trailing-newline":
            _action_remove_nl(item, env, ctx, output_dir)
        else:
            assert False
    return tpl


def process(tpl, env, ctx, **kwargs):
    output_dir = kwargs["output_dir"]

    items = reversed(tpl.includes)
    for item in items:
        _process(item, env, ctx, output_dir)

    return tpl



#!/usr/bin/env python3
# pylint: disable=redefined-outer-name
#
import os
import shutil
import subprocess
import tempfile
import typing as t

from setuptools import setup

try:
    import tomlib
except ModuleNotFoundError:
    try:
        import tomllib as tomlib
    except ModuleNotFoundError:
        import tomli as tomlib


class AutoRestoreFile:
    def __init__(self, src: str, dst: str):
        self.src = src
        self.dst = dst

    def __enter__(self) -> None:
        shutil.copyfile(self.src, self.dst)

    def __exit__(self, _exc_type: t.Any, _exc_value: t.Any, _exc_traceback: t.Any) -> None:
        shutil.copyfile(self.dst, self.src)


def load_requirements() -> list[str]:
    with open("requirements.txt", encoding="UTF-8") as fd:
        requirements = fd.read().splitlines()
        requirements = [x for x in requirements if x and not x.startswith("--")]
    return requirements


def save_requirements(requirements: list[str]) -> None:
    with open("requirements.txt", mode="w", encoding="UTF-8") as fd:
        fd.writelines([f"{x}\n" for x in requirements])


def push_requirements() -> None:
    # requirements.txt support flags such as "--index", remove
    # them before we continue
    requirements = load_requirements()
    save_requirements(requirements)


def freeze_requirements() -> None:
    # If pyproject specifies creating script files, then
    # compile a static version of the dependencies.

    # Does the pyproject.toml have a "project.scripts"?
    config = os.path.dirname(__file__) + "/pyproject.toml"
    with open(config, mode="rb") as fd:
        pyproject = tomlib.load(fd)

    if not pyproject["project"].get("scripts"):
        return

    cmdline = "uv pip compile requirements.txt --format requirements.txt -o requirements.txt"
    retval = subprocess.run(cmdline, shell=True, check=True, capture_output=True)
    retval.check_returncode()

    # Make a copy as a `constraints.txt` file so it can be used optionally for dev
    shutil.copyfile("requirements.txt", "constraints.txt")


with tempfile.NamedTemporaryFile() as tmpfile:
    with AutoRestoreFile("requirements.txt", tmpfile.name):
        push_requirements()
        freeze_requirements()
        setup()



import os
import shutil
import sys

import jinja2
from boltons import fileutils


# TODO: properly copy all modes and stats
#
def _clone_file(env, ctx, src, dst):
    try:
        with open(src, encoding="UTF-8") as fdr, open(dst, mode="w", encoding="UTF-8") as fdw:
            dst_contents = fdr.read()
            dst_contents = env.from_string(dst_contents).render(**ctx)
            fdw.write(str(dst_contents or ""))
        shutil.copymode(src, dst)

    except UnicodeDecodeError:
        with open(src, mode="rb") as fdr, open(dst, mode="wb") as fdw:
            dst_contents = fdr.read()
            fdw.write(dst_contents)
        shutil.copystat(src, dst)
        shutil.copymode(src, dst)

    except jinja2.exceptions.UndefinedError as err:
        sys.exit(f'error: while rendering "{src}", encountered template error: {err}')
    except jinja2.exceptions.TemplateSyntaxError as err:
        sys.exit(f'error: while rendering "{src}", encountered template error: {err}')


def _clone(tpl, env, ctx, output_dir):
    def templatize(src, string):
        # Change templates in file name
        try:
            return str(env.from_string(string).render(**ctx))
        except jinja2.exceptions.UndefinedError as err:
            sys.exit(f'error: templating error during "{src}" file: {err}')

    if not os.path.isdir(output_dir):
        fileutils.mkdir_p(output_dir)

    src_path = tpl.workpath
    src_path = os.path.join(src_path, "template")
    src_path = os.path.abspath(src_path)

    if not os.path.isdir(src_path):
        print(
            'warning: no "template" named folder found in source template directory. No templating can be done',
            file=sys.stderr,
        )

    for src in fileutils.iter_find_files(src_path, "*", include_dirs=True):
        dst = os.path.relpath(src, src_path)
        dst = templatize(src, dst)
        dst = os.path.join(output_dir, dst)

        if os.path.islink(src):
            # Read the target, and if templatized target is the same
            # as original target, then we can just copy the link over
            target_src = os.readlink(src)
            target_dst = templatize(src, target_src)
            if target_src == target_dst:
                if os.path.exists(dst):
                    os.unlink(dst)
                shutil.copyfile(src, dst, follow_symlinks=False)
            else:
                # TODO: make a new link in dst that targets a
                # templatized value
                sys.exit("error: changing link is not implemented while handling {src}")
        elif os.path.isdir(src):
            fileutils.mkdir_p(dst)
        elif os.path.isfile(src):
            fileutils.mkdir_p(os.path.dirname(dst))
            _clone_file(env, ctx, src, dst)
        else:
            sys.exit(f'error: path copy failed. File "{src}" of unknown type')

    return tpl, ctx


def clone(tpl, env, ctx, **kwargs):
    output_dir = kwargs["output_dir"]

    items = reversed(tpl.includes)
    for item in items:
        _clone(item, env, ctx, output_dir)

    return tpl



from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    """
    Return { value to use, stop (True/False) }

    If stop is True, the rest of the validators are skipped.
    """
    # ctrl-d gives us a None, if value is not None,
    # we're not concerned
    if value is not None:
        return value, False

    nullable = schema.get("nullable", False)
    if nullable:
        return None, True
    raise ValidationError(message="Answer is required")



# vim: foldmarker=[[,]] foldmethod=marker
import cerberus

# https://docs.python-cerberus.org/en/stable/validation-rules.html
SCHEMA = r"""
---
actions: # [[
  type: list
  default: []
  schema:
    type: dict
    default: {}
    nullable: true
    schema:
      # actions.hidden
      # --------------
      action:
        type: string
        required: True
        allowed: ["move", "remove", "remove-trailing-newline"]

      # actions.dst
      # -----------
      dst:
        type: string
        excludes: ["path"]

      # actions.else
      # ------------
      else:
        type: string
        allowed: ["remove"]

      # actions.if
      # ----------
      if:
        type: [string, boolean]
        required: True

      # actions.order
      # -------------
      order:
        type: integer
        default: 0

      # actions.path
      # ------------
      path:
        type: [string, list]
        excludes: ["dst", "src"]

      # actions.src
      # -----------
      src:
        type: string
        excludes: ["path"]
# ]]

answers: # [[
  type: dict
  default: {}
# ]]

inherit: # [[
  type: list
  default: []
  schema:
    type: dict
    default: {}
    nullable: true

    schema:
      # inherit.branch
      # --------------
      branch:
        type: string
        nullable: True

      # inherit.include
      # ---------------
      include:
        type: string
        required: true

      # inherit.filename
      # ---------------
      filename:
        type: string
        default: "scaffold.yml"
# ]]

jinja2: # [[
  type: dict
  default: {}
  schema:
    # lstrip_blocks
    lstrip_blocks:
      type: boolean
      default: false

    # trim_blocks
    trim_blocks:
      type: boolean
      default: false
# ]]

questions: # [[
  type: list
  default: []
  schema:
    type: dict
    default: {}
    nullable: true
    schema:

      # questions.description
      # ---------------------
      description:
        type: string

      # questions.hidden
      # ----------------
      hidden:
        type: [string, boolean]
        default: false

      # questions.if
      # ------------
      if:
        type: [string, boolean]

      # questions.name
      # --------------
      name:
        type: string
        regex: '^\S+$'
        required: true
        minlength: 1

      # questions.order
      # ---------------
      order:
        type: integer
        default: 0

      # questions.schema
      # ----------------
      schema:
        type: dict
        nullable: true
        coerce: asdict
        check_with: schema_rules
        schema:

          # questions.schema.allowed
          # ------------------------
          allowed:
            type: list
            schema:
              type: [string, integer]

          # questions.schema.default
          # ------------------------
          default:
            type: [string, boolean, integer]

          # questions.schema.nullable
          # -------------------------
          nullable:
            type: boolean
            default: False

          # questions.schema.max_length
          # ---------------------------
          max_length:
            type: integer

          # questions.schema.max_value
          # ---------------------------
          max_value:
            type: integer

          # questions.schema.min_length
          # ---------------------------
          min_length:
            type: integer
            min: 1

          # questions.schema.min_value
          # ---------------------------
          min_value:
            type: integer

          # questions.schema.type
          # ---------------------
          type:
            type: string
            default: "string"
            allowed: ["string", "integer", "boolean"]
# ]]
"""


class LocalValidator(cerberus.Validator):
    def _normalize_coerce_asdict(self, value):
        if value is None:
            return {}
        return value

    def _normalize_coerce_tolist(self, value):
        if isinstance(value, str):
            return [value]
        return value

    def _check_with_schema_rules(self, field, schema):
        pass


#    # TODO: Validations
#    def _check_with_include_minlength(self, field, value):
#        if value is None:
#            self._error(field, "Length of an 'include' path must be non-zero")
#            return
#
#        for include in value:
#            if len(include) == 0:
#                self._error(field, "Length of an 'include' path must be non-zero")
#                return
#
#    def _check_with_schema_rules(self, field, schema):
#        if schema.get("type") in ["boolean", "string"] and any(k in schema for k in ["min", "max"]):
#            self._error(field, 'neither "min" nor "max" can be specified if "type" is either boolean or string')
#            return
#        if schema.get("type") in ["boolean", "integer"] and any(k in schema for k in ["min_length", "max_length"]):
#            # fmt: off
#            self._error(field, 'neither "min_length" nor "max_length" can be specified if "type" is either boolean or integer')
#            return
#            # fmt: on
#        if "allowed" in schema and schema.get("type") == "boolean":
#            self._error(field, '"allowed" can not be specified if "type" is boolean')
#            return
#        if "allowed" in schema and any(k in schema for k in ["min_length", "max_length", "min", "max"]):
#            # fmt: off
#            self._error(field, '"allowed" can not be specified when one of "min_length", "max_length", "min", or "max" is also specified')
#            return
#            # fmt: on
#        if "min" in schema and "max" in schema:
#            if int(schema["min"]) > int(schema["max"]):
#                self._error(field, '"min" must be inferior to "max"')
#                return
#        if "min_length" in schema and "max_length" in schema:
#            if schema["min_length"] > schema["max_length"]:
#                self._error(field, '"min_length" must be inferior to "max_length"')
#                return
#
#    def _check_with_file_rules(self, field, value):
#        if "else" in value and "if" not in value:
#            self._error(field, '"if" required if a "files" element contains an "else"')
#            return
#        if "move" in [value.get("action", ""), value.get("else", "")] and "dest" not in value:
#            self._error(field, '"dest" required if a "files" element contains an "action" or "else"')
#            return
